/*
 * vlcrc.c - Control VLC program by unix socket
 *
 * Copyright (C) 2012-2013  Davide Parise <bubini.mara5@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 * Start execution of an instance the vlc programm by rc control.
 * Using a local socket to send and read command from and to vlc.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <errno.h>

#include <sys/socket.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/un.h>

#include "../constant.h"
#include "../file_browser/filebrowser.h"
#include "../util/sympleLinkedList.h"

#include "vlcrc.h"
#include "command.h"

/* file extension that can open vlc */
const char *multimedia_file_extension[] = { 
		/* video */
		".asf", ".avi", ".divx", ".dv", ".flv", ".gxf", ".m1v", ".m2v", ".m2ts", ".m4v", 
		".mkv", ".mov", ".mp2", ".mp4", ".mpeg", ".mpeg1", ".mpeg2", ".mpeg4", ".mpg", 
		".mts", ".mxf", ".ogg", ".ogm", ".ps", ".ts", ".vob", ".wmv",
		/* audio */		
		".a52", ".aac", ".ac3", ".dts", ".flac", ".m4a", ".m4p", ".mka", ".mod",
		".mp1", ".mp2", ".mp3", ".ogg", ".oma", ".spx", ".wav", ".wma", ".xm",
		/* playlist */
		".b4s", ".m3u", ".pls", ".vlc", ".xspdf",
		/* other here */

		(char*)NULL
		};

char *ARGV_VLC[6];	/* argument */
char *file_socket;	/* file socket */

int sock;	/* local socket for send command to vlc */
struct list_t *playlist;		/* item in the playlist */
int isStartVlc;		/* 0(TRUE) if there is an instance of vlc 1(FLASE) otherwise */



/* MOVE IT IN UTIL */

/* 
 * set socketd fd in blocking mode
*/
int setBlocking(int fd){
    int flags;

    /* If they have O_NONBLOCK, use the Posix way to do it */
#if defined(O_NONBLOCK)
    /* Fixme: O_NONBLOCK is defined but broken on SunOS 4.1.x and AIX 3.2.5. */
    if (-1 == (flags = fcntl(fd, F_GETFL, 0)))
        flags = 0;
    return fcntl(fd, F_SETFL, flags & (~O_NONBLOCK));
#else
    /* Otherwise, use the old way of doing it */
    flags = 0;
    return ioctl(fd, FIOBIO, &flags);
#endif

}

/*
 * set socket fd in non blocking mode 
*/
int setNonblocking(int fd)
{
    int flags;

    /* If they have O_NONBLOCK, use the Posix way to do it */
#if defined(O_NONBLOCK)
    /* Fixme: O_NONBLOCK is defined but broken on SunOS 4.1.x and AIX 3.2.5. */
    if (-1 == (flags = fcntl(fd, F_GETFL, 0)))
        flags = 0;
    return fcntl(fd, F_SETFL, flags | O_NONBLOCK);
#else
    /* Otherwise, use the old way of doing it */
    flags = 1;
    return ioctl(fd, FIOBIO, &flags);
#endif
}     

/* 
 * operation on the socket 
 * clear buffer in input (read) of the local socket 
*/
int clear_sock_in(){
	char tmp;
	int er;

	setNonblocking(sock);
	while((er=read(sock,&tmp,1)) == 1);
	if(!((errno == EAGAIN) || (errno == EWOULDBLOCK))){
		perror("Error:clear_sock_in() sock ");
		return -1;
	}
	setBlocking(sock);
	return 0;
}



/* UTIL PLAYER */

/*
 * add or enqueue the file name
 * if file is directory then execute command to all files contenents in this directory
 * use filebrowser.c for change contenent of directories
 * @Param
 *	name: the name of file or directory conteneted in current directory of filebrowser
 *		if name is file the command is executed 
 *		if the name is a directory the command is executed for all file in the directory
 *	command: must do one of VLCRC_CMD_ADD for add element or 
 *		VLCRC_CMD_ENQUEUE for enqueue element
*/
int compute(char *name,int command){
	char tmp[MAX_PATH_LENGTH];
	struct stat data;
	int(*f)(char*);

	/* get function by command */
	switch(command){
		case VLCRC_CMD_ADD:
			f = addFile;
			break;
		case VLCRC_CMD_ENQUEUE:
			f = enqueueFile;
			break;
		default:
			fprintf(stderr,"Unknow Command in compute\n");
			return -1;
	};

	/* NOT CHECK PATH LENTGTH 
	 possibility of buffer overflow 
	*/
	memset(tmp,'\0',MAX_PATH_LENGTH);
	strncpy(tmp,getCurrentPath(),strlen(getCurrentPath()));
	strncat(tmp,"/",1);
	strncat(tmp,name,strlen(name));

	stat(tmp,&data);
	if(S_ISREG(data.st_mode)){

		f(tmp);
	}
	else if(S_ISDIR(data.st_mode)) {

		computeRecursiveDir(tmp,f);
	}
	else{
		fprintf(stderr,"Error compute player command:file type not know %s\n",tmp);
		return -1;
	}
	/* other here */
	return 1;
}

/* read line from the socket sock 
 * @line: the result string
 * @return: number of character readed 
*/
int read_line_c(char **line){
	int nrecv,i=0;
	char *buf;
	buf = malloc(sizeof(char)*MAX_PATH_LENGTH);
	if(buf == NULL){
		perror("read_line_c: error while alloc");
		return -1;
	}

	for(;;){
		nrecv = read(sock,&buf[i],1);
		if(nrecv == -1){
			perror("read_line_c: read socket");
			return -1;
		}
		if(nrecv == 0){
			break;
		}

		if(buf[i] == '\n'){
			break;
		}

		i++;

		/* modify controllo whit realloc for example */
		if(i>MAX_PATH_LENGTH){
			void* ptr = realloc(buf,sizeof(char)*(MAX_PATH_LENGTH*2));
			if(ptr==NULL){
				perror("read_line_c:realloc not space in buffer");
				return -1;
			}
		}	
	}
	if(i>0){
		*line=malloc(sizeof(char*)*i);
		if(*line==NULL){
			perror("readline_c: error malloc");
			return -1;
		}
		strncpy(*line,buf,i);
	}
	return i;	/* len of byte readed */
}

/*
 * send command to the instance of vlc by socket operations
 * the argument are command to execute and param is argument of command
*/
int send_command_c(char *command,char *param){
	int nwrite;
	char *end = "\n";
	size_t count;
	if(isStartVlc == 0){	/* vlc is not executing */
		fprintf(stderr,"Vlcrc:Error Send command to vlc\nThere is not an instance of vlc\n");
		return -1;
	}
	if(command == NULL)
		return -1;

	count = sizeof(char)*strlen(command);
	nwrite = write(sock,command,count);
	if(nwrite<0){
		perror("Send_command_c(): can not write in the socket\n");
		return -1;
	}
	if(param != NULL){
		count = sizeof(char)*strlen(param);
		nwrite += write(sock,param,count);
		if(nwrite<0){
			perror("Send_command_c(): can not write in the socket\n");
			return -1;
		}
	}
	nwrite += write(sock,end,1);
	if(nwrite<0){
		perror("Error send_command_c(): can not write in the socket\n");
		return -1;
	}

	clear_sock_in();
	
	return nwrite;

}

/* PLAYER */

/*
 * initialize data and structures
*/

void init_vlcrc(){
	int len=0;
	char *lbcs;
	char *socket_filename = "/socket";
	printf("Init Vlc Controller:set filebrowser and create empty playlist\n");
	playlist = createList();
	init_filebrowser();
	isStartVlc=0;

	lbcs = getCurrentPathLBCS();
	len = strlen(lbcs) + strlen(socket_filename);
	file_socket = malloc(sizeof(char)*(len+1));
	if(file_socket==NULL){
		perror("init_vlcrc: file_socket");
		return;
	}
	memset(file_socket,'\0',len+1);
	strncpy(file_socket,lbcs,strlen(lbcs));
	strncat(file_socket,socket_filename,strlen(socket_filename));
	printf("Init Vlc Controller:Set local socket %s\n",file_socket);

}

/*
 * Free all resources
*/
void free_vlcrc(){
	if(isStartVlc==1)
		quit_vlcrc();

	if(file_socket != NULL)
		free(file_socket);

	if(playlist != NULL)
	deleteList(playlist,free);

}
int start_vlcrc(){
	int pid,action;
	int tryConnect=10;	/* try 10 times to connect */	
	struct sockaddr_un addr;

	clearListAndFree(playlist,free);
	memset((void *)&addr,0,sizeof(addr));
	addr.sun_family = AF_UNIX;
	strcpy(addr.sun_path,file_socket);

	sock = socket(PF_LOCAL,SOCK_STREAM,0);
	if(sock<0){
		perror("Impossibile aprire socket\n");
		return -1;
	}
	isStartVlc = 1;
	pid = fork();
	if(pid<0){
		perror("Errore fork\n");
		return -1;
	}
	else if(pid == 0){
		printf("run vlc remote controll... \n");

#ifdef HOLD_VLC 
		int r = execlp("vlc","vlc","--show-intf","--extraintf",\
			"oldrc","--rc-unix",file_socket,(char*)NULL);
#else
		int r = execlp("vlc","vlc","--extraintf","oldrc", \
			"--rc-unix",file_socket,(char*)NULL);
#endif
		if(r==-1){
			perror("PROCESSO FIGLIO ERRORE\n");
			isStartVlc = 0;
			return -1;
		}
		/* never here! Cant force close socket!! */
		exit(0);	
	}
	else{
		while(tryConnect--){
			sleep(3);	
			if(connect(sock,(struct sockaddr*)&addr,sizeof(addr))<0){
				if(errno == ENOENT){
					/* file not exist */
					printf("Start Vlcrc reconnect time %d\n",tryConnect);
				}
				else{			
					/* error */
					perror("Start Vlcrc:Can not connect to the socket\n");
					wait(&action);
					return -1;
				}
			}
			else{
				printf("\nOK\nNormal execution:: accept command\n");
				return 0;	
			}
		}
		isStartVlc=0;
		return -1;
	}
}

int quit_vlcrc(){
	int child;

	send_command_c("quit",NULL);

	printf("Wait VLC quit \n");
	wait(&child);
	if(child == -1){
		perror("Quit Vlc \n");
		isStartVlc=1;
		return -1;
	}
	close(sock);

	isStartVlc=0;

	return 0;

}
int getIsStartVlcrc(){
	return isStartVlc;
}
/* 
 * if there is an instance of vlc running then quit it
 * otherwise create one  
 */
int toggle_vlcrc(){
	if(isStartVlc == 0){
		return start_vlcrc();
	}
	return quit_vlcrc();
}

/************/
/* playlist */
/************/

/* ???
 * modified at last time
 * i can not read the playlist whit the socket because the output is not regular(standard)
 * i have added a new internal list of item and added new functions for manipulate it
 * (but i do not like this)
 * i must rebuild all
*/


/*
 * Check if the file name passed by argument is a multimedia file.
 * It no find the mime type but do it by compare the extension.
 * @param:
 *	filename: The name of the file
 * @return:
 *	1 - if it is a multimedia file
 *	0 - otherwise
*/
int isFileMultimedia(char *filename){
	register int i;
	int lenfile,lenext;

	lenfile = strlen(filename);
	for(i=0;multimedia_file_extension[i] != NULL;i++){
		lenext = strlen(multimedia_file_extension[i]);
		if(lenfile>lenext){
			if(strncmp(&filename[lenfile-lenext],multimedia_file_extension[i],lenext) == 0)// compare last char
				return 1;
		}
	}
	return 0;
}

/* where is stored current playlist */
struct list_t *getPlaylist(){
	return playlist;
}

/* print on the standard output the elements of playlist */
void printPlaylist(){
	struct item_t *iter;
	char *tmp;

	printf("\n-------PLAYLIST-------\n");
	iter = getIterator(getPlaylist());
	while(iter!=NULL){
		tmp = (char*)getDataItem(iter);
		if(tmp!=NULL)
			printf("%s\n",tmp);
		else
			printf("Warning:No data file in item\n");

		iter = getNextItem(iter);
	}
	printf("\n-------END PLAYLIST-------\n");
}


/*
 * enqueue a file in at the end of playlist
*/
int enqueueFile(char *name){
	char *new_name;
	int len_name;

	if(isFileMultimedia(name)==0)	/* if it is not a multimedia file, do nothing */
		return 0;
	/* 
	* copy the name whit path for the internal playlist 
	* the argument name is always a local variable 
	* alloc new space and free when is unused
	*/

	len_name = strlen(name);
	 
	new_name = malloc(sizeof(char)*(len_name+1));
	if(new_name == NULL){
		perror("enqueueFile() malloc error");
		return -1;
	}
	memset(new_name,'\0',len_name+1);
	strncpy(new_name,name,len_name);
	addItemList(playlist,(void *)new_name);

	/* real enqueue */
	if(send_command_c("enqueue ",name))
		return 1;

	fprintf(stderr,"player:Error enqueue File\n");
	return -1;
}

/*
 * enqueue a file in at the end of playlist and play it
*/
int addFile(char *name){
	char *new_name;
	int len_name;

	if(isFileMultimedia(name)==0)	/* if it is not a multimedia file, do nothing */
		return 0;

	/* copy the name for the internal playlist (the same as enqueue())*/
	len_name = strlen(name);
	 
	new_name = malloc(sizeof(char)*(len_name+1));	/* length more null-character */
	if(new_name == NULL){
		perror("addFile() malloc error");
		return -1;
	}
	memset(new_name,'\0',len_name+1);
	strncpy(new_name,name,len_name);

	addItemList(playlist,(void *)new_name);

	if(send_command_c("add ",name))
		return 1;

	fprintf(stderr,"player:Error add File\n");
	return -1;
}

/* enqueue all file contenent in directory name */
int enqueueDir(char *name){
	return computeRecursiveDir(name,enqueueFile);
}
/* add all files content in the directory name */
int addDir(char *name){
	return computeRecursiveDir(name,addFile);
}

/* clear all elements of the playlist */
int clearPlst(){
	clearListAndFree(getPlaylist(),free);
	return send_command_c("clear",NULL);
}
/* play next item in the playlist */
int next(){
	return send_command_c("next",NULL);
}
/* play previus item in the playlist */
int prev(){
	return send_command_c("prev",NULL);
}
/* play elemnt at position index of the playlist */
int gotoPlstS(char *index){
	return send_command_c("goto ",index);
}
/* goto at element at position index of playlist */
int gotoPlstI(int index){
	char *a;
	a = malloc(sizeof(int));
	if(a == NULL){
		perror("Vlcrc:gotoPlstI\n");
		return -1;
	}
	sprintf(a,"%d",index);

	return send_command_c("goto ",a);
}

/* execution */
int play(){
	return send_command_c("play",NULL);
}
int pause(){
	return send_command_c("pause",NULL);
}
int isPlaying(){
	char *line;
	int result;

	clear_sock_in();
	send_command_c("is_playing ",NULL);
	read_line_c(&line);

	sscanf(line,"%d",&result);
	free(line);

	return result;
}
int stop(){
	return send_command_c("stop",NULL);
}
int seek(char *sec){
	return send_command_c("seek ",sec);
}

int jumpI(int sec){
	char command[20];
	char *tmp;
	int time;

	clear_sock_in();
	send_command_c("get_time",NULL);
	read_line_c(&tmp);
	sscanf(tmp,"%d",&time);

	time+=sec;

	sprintf(command,"seek %d",time);

	free(tmp);
	return send_command_c(command,NULL);
}

int getTime(char *sec){
	if(isPlaying()){
		clear_sock_in();
		send_command_c("get_time",NULL);
		return read_line_c(&sec);
	}
	else
		return -1;
}
int getLength(char *len){
	if(isPlaying()){
		clear_sock_in();
		send_command_c("get_length ",NULL);
		return read_line_c(&len);
	}
	else
		return -1;
}
int volup(char *step){
	return send_command_c("volup ",step);
}
int voldown(char *step){
	return send_command_c("voldown ",step);	
}
int volmute(){
	return send_command_c("volume ","0");
}
int fullscreen(char *state){
	return send_command_c("f",state);
}

