/*
 * msg.c - Define and manipulate the messages structure
 *
 * Copyright (C) 2012-2013  Davide Parise <bubini.mara5@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 *	This is the message send by socket (now bluetooth)
 *
 *	/*----------------------------- HDR(8B) --------------------|--DATA(nB)---*
 *	/* service_id(1B) | flags(1B) | type(2B) | size of data(4B) | data(nB) ---* 
 *	/*----- HDR --------------------------------------------------------------*
 *		@service_id: is the identifier of the application/client that send the message
 *		@type: is the type of message, possible type are command, request, ack, error, and other
 *		@data: not only messages have data but if there is data is add at the end
 *
 *
 *	changes whit old protocol:
 *
 *	1) STANDARD FORMAT MESSAGE (the header is the same)
 *		contro: adding hoveread
 *		pro:	simplify manipulation the messages
 *	2) FUNCTION FOR MANIPULATE A STRUCTURE THAT REPRESENT A MESSAGE (get byte,create msg,...)
 *		contro:	increase exectution time and memory usage
 *		pro:	clarity, same operation for all type of messages
 *	3)MESSAGE COMUNICATION PROTOCOL:
 *		even ack
 *		handler error
 *
 *	Definitions about byte order for mapping type:
 *	the messages byte order is Big-endian (the msb is the base of the pointer)
*/

#ifndef MSG_H
#define MSG_H

#include <stdint.h>

#define MAX_MESSAGE_SIZE	256	/* ? TEMPORANY MAX IS ~4G */
#define MSG_HEADER_SIZE		8

typedef uint8_t msg_t;	/* this structure only map an array of byte */


/*************************************************/
/*************** Error "Handling" ****************/
/*************************************************/
const char *strmsgerr(int err);

/**********************************************/
/************** MSG FUNCTION HDR **************/
/**********************************************/
int getMsgLength(msg_t *msg,uint32_t *size); /* data + header */
int setMsgServiceId(msg_t *msg,uint8_t service_id);
int getMsgServiceId(msg_t *msg,uint8_t *service_id);
int getMsgFlags(msg_t *msg,uint8_t *flags);
int setMsgType(msg_t *msg,uint16_t type);
int getMsgType(msg_t *msg,uint16_t *type);
int setMsgTypeM(msg_t *msg,uint8_t type);
int getMsgTypeM(msg_t *msg,uint8_t *type);
int setMsgTypeL(msg_t *msg,uint8_t type);
int getMsgTypeL(msg_t *msg,uint8_t *type);
int setMsgSize(msg_t *msg,uint32_t size);
int getMsgSize(msg_t *msg,uint32_t *size);

/**********************************************/
/************* MSG  DATA FUNCTION *************/
/**********************************************/

/* add data to exsisting message */
int addMsgData(msg_t *msg,uint8_t *data,uint32_t size);

/* get a pointer to data */
uint8_t *getPtrMsgData(msg_t *msg);

/**********************************************/
/**************** MSG FUNCTIONS ***************/
/**********************************************/

msg_t *createMsg(uint8_t service_id,uint8_t flags,uint8_t typeM,uint8_t typeL,uint32_t size,uint8_t *data);
msg_t *createMsgI(uint8_t service_id,uint8_t flags,uint8_t typeM,uint8_t typeL,int data);
msg_t *createMsgC(uint8_t service_id,uint8_t flags,uint8_t typeM,uint8_t typeL,char *data);
msg_t *createMsgV(uint8_t service_id,uint8_t flags,uint8_t typeM,uint8_t typeL,const char *data_type,...);
void freeMsg(msg_t *msg);

/**********************************************/
/************ UTILITY MSG FUNCTIONS ***********/
/**********************************************/

/*
 * utility function for convert a buff into a messages format 
 * util for example when call read whit a local variable
*/
int buffToMsg(msg_t **msg,uint8_t *buf,int len);

//int msgToBuff(msg_t *msg,uint8_t **buf);

/* printing type format */
#define PRINT_MSG_NO_DATA	0
#define PRINT_MSG_BYTE_DATA	1
#define PRINT_MSG_CHAR_DATA	2
#define PRINT_MSG_STR_DATA	3
/*
 * utility function for debuggin 
 * print a message header and data 
*/
void printMsgStruct(msg_t *msg,int format);

/*************************************************/
/***************** Internal Queue ****************/
/*************************************************/
void init_queue();
void enqueue(uint8_t *data);
uint8_t *dequeue();
void clear_queue();		/* clear all element in the queue but not data */
void remove_data_queue();	/* clear all element and data in the queue */
int size_queue();

#endif

