/*
 * msg.c - Define and manipulate the messages structure
 *
 * Copyright (C) 2012-2013  Davide Parise <bubini.mara5@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 *	This is the message send by socket (now bluetooth)
 *
 *	/*----------------------------- HDR(8B) --------------------|--DATA(nB)---*
 *	/* service_id(1B) | flags(1B) | type(2B) | size of data(4B) | data(nB) ---* 
 *	/*----- HDR --------------------------------------------------------------*
 *		@service_id: is the identifier of the application/client that send the message
 *		@type: is the type of message, possible type are command, request, ack, error, and other
 *		@data: not only messages have data but if there is data is add at the end
 *
 *
 *	changes whit old protocol:
 *
 *	1) STANDARD FORMAT MESSAGE (the header is the same)
 *		contro: adding hoveread
 *		pro:	simplify manipulation the messages
 *	2) FUNCTION FOR MANIPULATE A STRUCTURE THAT REPRESENT A MESSAGE (get byte,create msg,...)
 *		contro:	increase exectution time and memory usage
 *		pro:	clarity, same operation for all type of messages
 *	3)MESSAGE COMUNICATION PROTOCOL:
 *		even ack
 *		handler error
 *
 *	Definitions about byte order for mapping type:
 *	the messages byte order is Big-endian (the msb is the base of the pointer)
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <assert.h>
#include <errno.h>

#include "h/msg.h"

#define OFFSET_SERVICE_ID	0
#define OFFSET_FLAGS		1
#define OFFSET_TYPE		2
#define OFFSET_SIZE		4
#define OFFSET_DATA		8

const char DATA_TYPE_8 = '1';	/* 1 byte */
const char DATA_TYPE_16 = '2';	/* 2 byte */
const char DATA_TYPE_32 = '4';	/* 4 byte */



/*************************************************/
/**************** Byte Convert  ******************/
/*************************************************/

int getInt32(uint8_t *buf){
	int result=0;

#if __BYTE_ORDER == __BIG_ENDIAN
	result = (int) ( (buf[3] << 24) | (buf[2]<<16) | (buf[1]<<8) | buf[0]);
#elif __BYTE_ORDER == __LITTLE_ENDIAN
	result = (int) ( (buf[0] << 24) | (buf[1]<<16) | (buf[2]<<8) | buf[3]);
#else
	printf("unknow type:Can not convert type\n");
	return 0;
#endif
	
	return result;
}

void setInt32(uint8_t *msg,uint32_t data){
#if __BYTE_ORDER == __BIG_ENDIAN
	msg[3] = (uint8_t) data >> 24;
	msg[2] = (uint8_t) data >> 16;
	msg[1] = (uint8_t) data >> 8;
	msg[0] = (uint8_t) data;
#elif __BYTE_ORDER == __LITTLE_ENDIAN
	msg[0] = (uint8_t) data >> 24;
	msg[1] = (uint8_t) data >> 16;
	msg[2] = (uint8_t) data >> 8;
	msg[3] = (uint8_t) data;
#else
	printf("unknow type:Can not convert type\n");
#endif
}

void setInt16(uint8_t *msg,uint16_t data){
#if __BYTE_ORDER == __BIG_ENDIAN
	msg[1] = (uint8_t) (data >> 8);
	msg[0] = (uint8_t) data;
#elif __BYTE_ORDER == __LITTLE_ENDIAN
	msg[0] = (uint8_t) (data >> 8);
	msg[1] = (uint8_t) data;
#else
printf("unknow type:Can not convert type\n");
#endif
	//printf("Set 16: value %d 0x%.4x\n  (0x%.2x - 0x%.2x)\n",data,data,msg[0],msg[1]);
}

uint16_t getInt16(uint8_t *buf){
	uint16_t result;
#if __BYTE_ORDER == __BIG_ENDIAN
	result = (uint16_t) ((buf[1] << 8) | buf[0]);
#elif __BYTE_ORDER == __LITTLE_ENDIAN
	result = (uint16_t) ((buf[0] << 8) | buf[1]);
#else
	printf("unknow type\n");
	return 0;
#endif
	return result; 
}

/*************************************************/
/*************** Error "Handling" ****************/
/*************************************************/

int type_msg_err;

const char *str_msg_err[] = { 
	"No error",		/* 0 */
	"Unknow error type", 	/* 1 */
	"Error Header size",	/* 2 */
	"Malloc Fails", 	/* 3 */
	"Message is NULL", 	/* 4 */
	"Different Size",	/* 5 */
	"Message too large",	/* 6 */
	"Memory Error"		/* 7 */
};

void setmsgerr(int err){
	type_msg_err = err;
}

int getmsgerr(){
	int result;
	if(type_msg_err>0)
		result = -type_msg_err;
	else
		result = type_msg_err;

	return result;
}
const char *strmsgerr(int err){
	if(err<0)
		err = -err;
	return str_msg_err[err];
}

/**********************************************/
/************* MSG FUNCTION STRUCT ************/
/**********************************************/
int getMsgLength(msg_t *msg,uint32_t *size){
	if(msg==NULL){
		setmsgerr(4);
		return -4;
	}
	*size = getInt32(&msg[OFFSET_SIZE])+MSG_HEADER_SIZE;
	return 0;
}

int setMsgServiceId(msg_t *msg,uint8_t service_id){
	if(msg==NULL){
		setmsgerr(4);
		return -4;
	}
	msg[OFFSET_SERVICE_ID] = service_id;
	return 0;	
}
int getMsgServiceId(msg_t *msg,uint8_t *service_id){
	if(msg==NULL){
		setmsgerr(4);
		return -4;
	}
	*service_id = msg[OFFSET_SERVICE_ID];
	return 0;	

}

int setMsgFlags(msg_t *msg,uint8_t flags){
	if(msg==NULL){
		setmsgerr(4);
		return -4;
	}
	msg[OFFSET_FLAGS] = flags;
	return 0;
}
int getMsgFlags(msg_t *msg,uint8_t *flags){
	if(msg==NULL){
		setmsgerr(4);
		return -4;
	}
	*flags = msg[OFFSET_FLAGS];
	return 0;

}
int setMsgType(msg_t *msg,uint16_t type){
	if(msg==NULL){
		setmsgerr(4);
		return -4;
	}
	setInt16(&msg[OFFSET_TYPE],type);
	return 0;
}
int getMsgType(msg_t *msg,uint16_t *type){
	if(msg==NULL){
		setmsgerr(4);
		return -4;
	}
	*type = getInt16(&msg[OFFSET_TYPE]);
	return 0;
}

int setMsgTypeM(msg_t *msg,uint8_t type){
	if(msg==NULL){
		setmsgerr(4);
		return -4;
	}
	msg[OFFSET_TYPE] = type;
	return 0;
}
int getMsgTypeM(msg_t *msg,uint8_t *type){
	if(msg==NULL){
		setmsgerr(4);
		return -4;
	}
	*type = msg[OFFSET_TYPE];
	return 0;
}
int setMsgTypeL(msg_t *msg,uint8_t type){
	if(msg==NULL){
		setmsgerr(4);
		return -4;
	}
	msg[OFFSET_TYPE+1] = type;
	return 0;
}
int getMsgTypeL(msg_t *msg,uint8_t *type){
	if(msg==NULL){
		setmsgerr(4);
		return -4;
	}
	*type = msg[OFFSET_TYPE+1];
	return 0;
}

int setMsgSize(msg_t *msg,uint32_t size){
	if(msg==NULL){
		setmsgerr(4);
		return -4;
	}
	setInt32(&msg[OFFSET_SIZE],size);
	return 0;
}
int getMsgSize(msg_t *msg,uint32_t *size){
	if(msg==NULL){
		setmsgerr(4);
		return -4;
	}
	*size = getInt32(&msg[OFFSET_SIZE]);
	return 0;
}

/* 
 * add data to msg 
 * @param
 *	msg: a pointer to msg
 *	data: pointer to data
 *	size: size of new data express in byte
 * @result
 *	0 if ok <0 if an error occurs
*/
int addMsgData(msg_t *msg,uint8_t *data,uint32_t size){
	int old_size,new_size;
	msg_t *new_msg;

	if(msg==NULL){
		setmsgerr(4);
		return -4;
	}
	if(getMsgSize(msg,&old_size)<0){
		fprintf(stderr,"Error addMsgData() while add data: %s\n",strmsgerr(type_msg_err));
		return getmsgerr(); 
	}
	
	new_size = old_size+size;
	new_msg = realloc(msg,new_size+MSG_HEADER_SIZE);
	if(new_msg == NULL){
		fprintf(stderr,"Error addMsgData() while add data: %s\n",strerror(errno));
		setmsgerr(7);
		return getmsgerr();
	}
	/* set new size */
	setMsgSize(msg,new_size);
	/* copy data into new msg */
	memcpy(&msg[OFFSET_DATA+old_size],data,size);
	msg = new_msg;

	return 0;
}

/* 
 * return a pointer to the head of data 
 * @return: (ambiguo)
 *	if data exist return a pointer to data otherwise return NULL
 *	if an error occur return NULL and set error
*/
uint8_t *getPtrMsgData(msg_t *msg){
	int size;
	if(msg==NULL){
		setmsgerr(4);
		return NULL;
	}

	if(getMsgSize(msg,&size)<0){
		fprintf(stderr,"Error To get a pointer to msg data %s\n",strmsgerr(type_msg_err));
		return NULL;
	}
	setmsgerr(0);	/* no error */
	if(size>0)
		return &msg[OFFSET_DATA];
	else
		return NULL;	/* No Data */
}

/**********************************************/
/**************** MSG FUNCTIONS ***************/
/**********************************************/

/*
 * Create a new message whitouth data but only header
 * @param:
 * 	all header
 * @return:
 * 	a pointer to message or null if there is an error
*/
msg_t *createMsgHdr(uint8_t service_id,uint8_t flags,uint8_t typeM,uint8_t typeL){
	msg_t *msg;

	msg = malloc(sizeof(uint8_t)*(MSG_HEADER_SIZE));
	if(msg==NULL){
		perror("createMsg fails malloc raw");
		setmsgerr(3);
		return NULL;
	}
	setMsgServiceId(msg,service_id);
	setMsgFlags(msg,flags);
	setMsgTypeM(msg,typeM);
	setMsgTypeL(msg,typeL);
	setMsgSize(msg,0);

	return msg;
}

/*
 * Default implementation for Create a new message whit header and data passed as argument  
 * @param:
 *	all header 
 * 	size: the size express in byte of the data
 *	data: a pointer to data
 * @return:
 * 	a pointer to msg or null if are an error 
*/
msg_t *createMsg(uint8_t service_id,uint8_t flags,uint8_t typeM,uint8_t typeL,uint32_t size,uint8_t *data){
	msg_t *msg;

	if(data == NULL)
		msg = createMsgHdr(service_id,flags,typeM,typeL);
	else{	
		msg = malloc(sizeof(uint8_t)*(size+MSG_HEADER_SIZE));
		if(msg==NULL){
			perror("createMsg fails malloc raw");
			setmsgerr(3);
			return NULL;
		}
		setMsgServiceId(msg,service_id);
		setMsgFlags(msg,flags);
		setMsgTypeM(msg,typeM);
		setMsgTypeL(msg,typeL);
		setMsgSize(msg,size);
	
		memcpy(&msg[OFFSET_DATA],data,size);
	}

	return msg;
}

msg_t *createMsgI(uint8_t service_id,uint8_t flags,uint8_t typeM,uint8_t typeL,int data){
	int size = 4; //sizeof(uint32_t)
	msg_t *msg;
	
	msg = malloc(sizeof(uint8_t)*(size+MSG_HEADER_SIZE));
	if(msg==NULL){
		perror("createMsg fails malloc raw");
		setmsgerr(3);
		return NULL;
	}

	setMsgServiceId(msg,service_id);
	setMsgFlags(msg,flags);
	setMsgTypeM(msg,typeM);
	setMsgTypeL(msg,typeL);
	setMsgSize(msg,size);
	
	setInt32(&msg[OFFSET_DATA],data);

	return msg;
}

/* 
 * It is an utility, is the same as call createMsg()
 * whit param size = strlen(size) 
 * the param and result is the same as createMsg()
*/
msg_t *createMsgC(uint8_t service_id,uint8_t flags,uint8_t typeM,uint8_t typeL,char *data){
	if(data == NULL) /* not normal */
		return createMsgHdr(service_id,flags,typeM,typeL);
	else 
		return createMsg(service_id,flags,typeM,typeL,strlen(data),data);
}
/*
 * create a new message where data type is passed as argument whit *data_type param
 * and how many is derminate whit va_* util
 * @param:
 *	all header whitouth size
 *	data_type is a string to derminate the type of param passed
 *	... is the data
*/

msg_t *createMsgV(uint8_t service_id,uint8_t flags,uint8_t typeM,uint8_t typeL,const char *data_type,...){
	va_list Arg;			/* list of argument */
	va_start(Arg,data_type);	/* init list */
	register i;
	int len=0;
	uint8_t buf[MAX_MESSAGE_SIZE];	/* internal buffer */
	uint8_t *ptr;			/* pointer to a buffer */
	msg_t *msg;			/* result */

	ptr = &buf[OFFSET_DATA];
	for(i=0;data_type[i] != '\0';++i){
		if(data_type[i] == DATA_TYPE_8){
			*ptr = (uint8_t)va_arg(Arg,int);
			ptr++;len++;
		}
		else if(data_type[i] == DATA_TYPE_16){
			setInt16(ptr,(uint16_t)va_arg(Arg,int));
			ptr+=2;//sizeof(uint16_t)
			len+=2;
		}
		else if(data_type[i] == DATA_TYPE_32){
			setInt32(ptr,(uint32_t)va_arg(Arg,int));
			ptr+=4;
			len+=4;
		}
	}
	
	if(buffToMsg(&msg,buf,len+MSG_HEADER_SIZE)<0){
		printf("Error Can not convert message: %s\n",strmsgerr(type_msg_err));
		return NULL;
	}

	assert(msg!=NULL);

	setMsgServiceId(msg,service_id);
	setMsgFlags(msg,flags);
	setMsgTypeM(msg,typeM);
	setMsgTypeL(msg,typeL);
	setMsgSize(msg,len);
	
	return msg;
}

void freeMsg(msg_t *msg){
	if(msg != NULL)
		free(msg);
}
/*
 * utility function for convert a buff into a messages format 
 * util, for example, when call read whit a local buffer
*/
int buffToMsg(msg_t **msg,uint8_t *buf,int len){
	uint32_t size;

	/* check error */
	if(len < MSG_HEADER_SIZE){
		/* if len is less then header */
		setmsgerr(2);
		return -2;
	}
	if(len > MAX_MESSAGE_SIZE){
		/* if message is too large */
		setmsgerr(6);
		return -6;
	}

	*msg = malloc(sizeof(msg_t)*len);
	if(*msg==NULL){
		perror("Error malloc while convert buf to msg");
		setmsgerr(3);
		return -3;
	}

	memcpy(*msg,buf,len);

	return 0;
}

void printMsgStruct(msg_t *msg,int format){
	int i;
	uint8_t *data;
	uint32_t size;

	if(msg==NULL){
		printf("message is NULL\n");
	}
	else{
		getMsgSize(msg,&size);
		data = getPtrMsgData(msg);
		for(i=0;i<MSG_HEADER_SIZE;i++)
			printf("0x%.2x - ",msg[i]);
		printf("Data\n");
		switch(format){
			case PRINT_MSG_NO_DATA:
				break;
			case PRINT_MSG_BYTE_DATA:
				for(i=0;i<size;i++)
					printf("0x%.2x - ",data[i]);
				break;
			case PRINT_MSG_CHAR_DATA:	
				for(i=0;i<size;i++)
					printf("%c - ",data[i]);
				break;
			case PRINT_MSG_STR_DATA:
				printf("%s",data);
		}
		printf("\nEnd\n");
	}
	
}

/*************************************************/
/***************** Internal Queue ****************/
/*************************************************/

typedef struct list{
	struct list *next;
	uint8_t *data;
}List;

struct fifo{
	struct list *head;
	struct list *tail;
	int size;
}msg_fifo;


void init_queue(){
	msg_fifo.head = NULL;
	msg_fifo.tail = NULL;
	msg_fifo.size = 0;
}

void enqueue(uint8_t *data){
	struct list *msgl;

	msgl = malloc(sizeof(struct list));
	if(msgl == NULL){
		perror("add");
		return;
	}

	msgl->next = NULL;
	msgl->data = data;

	if(msg_fifo.head == NULL){
		msg_fifo.head = msgl;
		msg_fifo.tail = msgl;
	}
	else{
		((msg_fifo.tail)->next) = msgl;
		msg_fifo.tail = msgl;
	}
	msg_fifo.size++;
}

uint8_t *dequeue(){
	uint8_t *result = NULL;
	struct list *msgl;

	if(msg_fifo.head==NULL)
		return NULL;

	result = (msg_fifo.head)->data;
	msgl = msg_fifo.head;
	msg_fifo.head = (msg_fifo.head)->next;
	free(msgl);
	msg_fifo.size--;

	return result;
}
/* clear all element in the queue but not data */
void clear_queue(){
	while(dequeue() != NULL);
}
/* clear all element and data in the queue */
void remove_data_queue(){
	uint8_t *data;
	while((data=dequeue()) != NULL)
		free(data);
}
int size_queue(){
	return msg_fifo.size;
}


