
/* 
 * switcher.c - Core of protocoll comunication between server and client 
 * read messages and select service/application 
 *
 * Copyright (C) 2012-2013  Davide Parise <bubini.mara5@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <assert.h>

#include "h/constant_msg.h"
#include "h/switcher.h"
#include "h/io.h"
#include "h/mouse_handler.h"

#define ID_SERVER	0x00

#define CMD_QUIT	0XFF
#define CMD_SERVICE	0x03

msg_t ack_switcher[] = {
		ID_SERVER,	/* service id */
		FIELD_NOT_SET,	/* flags */
		TYPE_ACK,FIELD_NOT_SET,	/* type */
		0x00,0x00,0x00,0x00	/* size data = 0 */
	};


/***********************************************/
/************* Service Functions ***************/
/***********************************************/

#define MAX_SERVICE 256		/* 8bit */

struct service_t{
	uint8_t service_id;
	void (*handler)(void);
	char *name;
	char *desc;
};

struct service_t *curr_service[MAX_SERVICE];

/* initialize structures */
void init_app_service(){
	register i;
	for(i=0;i<MAX_SERVICE;i++)
		curr_service[i] = NULL;
}

/*
 * Register the apllications services
 * @param:
 *	service_id -> Unic id of the service
 *	name -> Name of the service
 *	desc -> A simply description
 *	handler -> Void function called when the service is request
 *
 * @return:
 *	0 -> If the service is registed
 *	-1 -> If an error occurs
 */
int register_app_service(uint8_t service_id,char *name,char *desc,void (*handler)){
	struct service_t *s;

	assert(service_id < MAX_SERVICE); 
	if(curr_service[service_id] != NULL){
		fprintf(stderr,"Error register_service(): conflict id \n");
		return -1;
	}

	s = malloc(sizeof(struct service_t));
	if(s==NULL){
		fprintf(stderr,"Error register_service() %s\n",strerror(errno));
		return -1;
	}

	s->service_id = service_id;
	s->handler = handler;
	if(name != NULL){
		s->name = malloc(sizeof(char)*(strlen(name)+1));
		memset(s->name,'\0',strlen(name)+1);
		if(s->name != NULL)
			strncpy(s->name,name,strlen(name));		
	}
	if(desc != NULL){
		s->desc = malloc(sizeof(char)*(strlen(desc)+1));
		memset(s->desc,'\0',strlen(desc)+1);
		if(s->desc != NULL)
			strncpy(s->desc,desc,strlen(desc));
	}
	/* other here */

	curr_service[service_id] = s;

	printf("Service register\n");
	print_app_service(service_id);

	return 0;
}

/* unregist the service whit id equal to service_id passed by argument */
void unregister_app_service(uint8_t service_id){

	if(curr_service[service_id] == NULL)
		return;	/* do nothing */
	if(curr_service[service_id]->name != NULL)
		free(curr_service[service_id]->name);
	if(curr_service[service_id]->desc != NULL)
		free(curr_service[service_id]->desc);

	free(curr_service[service_id]);
}

/* Unregister all service */
void unregister_apps_service(){
	register int i;
	
	for(i=0;i<MAX_SERVICE;i++)
		unregister_app_service(i);
}

/*
 * Print on the standard output the service whit id equal to service_id variable passed by argument
*/
void print_app_service(uint8_t service_id){
	if(curr_service[service_id] != NULL){
		printf("service_id: %d\n",service_id);
		if(curr_service[service_id]->name != NULL)
			printf("\tname: %s\n",curr_service[service_id]->name);
		else
			printf("\tname: no name\n");

		if(curr_service[service_id]->desc != NULL)
			printf("\tdesc: %s\n",curr_service[service_id]->desc);
		else
			printf("\tdesc: no desc\n");
	}
}

void print_all_app_service(){
	register int i;
	for(i=0;i<MAX_SERVICE;i++)
		print_app_service(i);
}

/*
 *	SWITCHER
*/

/*
 * Required by implementation
*/
void handler(){
	/* Do nothing */
}

/*
 * Initialize all structures
*/
void init_switcher(){
	init_app_service();
	/*register it self*/
	register_app_service(ID_SERVER,"Switcher","the main that switch all app",handler);
}

/*
 * Entering in loop
 * exit only when client close the connection or an error occurs
*/
int enter_loop(){
	msg_t *msg;
	struct service_t *service;
	uint8_t id,typeM,typeL;

	for(;;){
		printf("********* switcher ***************\n");
		if(read_msg(&msg)<0){
			if(errno == ECONNRESET){
				/* the connection is reset by peer */
				fprintf(stdout,"Connection closed by peer\n");
				return 0;
			}
			else{
				fprintf(stderr,"Error enter_loop() in server switcher while read msg\n");
				return -1;
			}
		}
		assert(msg!=NULL);

		if(getMsgServiceId(msg,&id)<0){
			fprintf(stderr,"Error loop switcher\n");
			freeMsg(msg);
			return -1;
		}
		if(getMsgTypeM(msg,&typeM)<0){
			fprintf(stderr,"Error loop switcher\n");
			freeMsg(msg);
			return -1;
		}
		if(getMsgTypeL(msg,&typeL)<0){
			fprintf(stderr,"Error loop switcher\n");
			freeMsg(msg);
			return -1;
		}

		if(typeM & TYPE_STATE_CHANGE){
			switch(typeL){
				case STATE_START:
					if(curr_service[id] == NULL){
						fprintf(stderr,"Error:Service not found\n");

						setMsgTypeM(msg,TYPE_ERROR);
						setMsgTypeL(msg,ERROR_SERVICE_NOT_AVIABLE);
						setMsgSize(msg,0);

						if(send_msg(msg)<0){
							fprintf(stderr,"Error loop_switcher() State start:can not send msg\n");
							return -1;
						}
						freeMsg(msg);
					}
					else{
						fprintf(stdout,"Start command received form:\n");
						print_app_service(id);

						setMsgTypeM(msg,TYPE_ACK);
						setMsgTypeL(msg,FIELD_NOT_SET);
						setMsgSize(msg,0);

						if(send_msg(msg)<0){
							fprintf(stderr,"Error loop_switcher() :can not send msg\n");
							return -1;
						}
						freeMsg(msg);
						/* goto controll to the client handler */
						(curr_service[id])->handler();
					}
					break;

				case STATE_STOP:/* NEVER HERE BUT IN THE OTHER HANDLER */
					printf("Stop command in the Server !!! (ID %d)\n",id);
					/* !! send ack !! */
					setMsgTypeM(msg,TYPE_ACK);
					setMsgTypeL(msg,FIELD_NOT_SET);
					setMsgSize(msg,0);
					if(send_msg(msg)<0){
						fprintf(stderr,"Error loop_switcher() State stop:can not send msg\n");
						freeMsg(msg);
						return -1;
					}
					freeMsg(msg);
					break;
				case STATE_RESUME:
					printf("Change state:ID %d Resume\n",id);
					/* send ack */
					setMsgTypeM(msg,TYPE_ACK);
					setMsgTypeL(msg,FIELD_NOT_SET);
					setMsgSize(msg,0);
					if(send_msg(msg)<0){
						fprintf(stderr,"Error loop_switcher() State resume:can not send msg\n");
						freeMsg(msg);					
						return -1;
					}
					freeMsg(msg);
					/* do anithing */
					(curr_service[id])->handler();				
					break;
				default:
					fprintf(stderr,"Error loop switcher::Change state ID %d:Unknow state 0x%.2x\n",id,typeL);
					setMsgTypeM(msg,TYPE_ERROR);
					setMsgTypeL(msg,ERROR_UNKNOW_COMMAND);
					setMsgSize(msg,0);
					if(send_msg(msg)<0){
						fprintf(stderr,"Error loop_switcher() Default state:can not send msg\n");
					}
					freeMsg(msg);
					/* do nothing */
					return -1;
				};

			}
			else{
				fprintf(stderr,"Error loop switcher::type command is not change state (ID:%d type 0x%.2x)\n",id,typeM);
				freeMsg(msg);
				return -1;
			}
		}
	}

/* starting loop */
void start_server(){
	printf("Server entering in loop\n");
	enter_loop();
	printf("Server exit loop\n");
	
}

